from collections import defaultdict
from operator import itemgetter
import re
from typing import Dict, Iterable, List, Optional, Tuple, Union

from .errors import MarkupError
from .style import Style
from .text import Span, Text


re_tags = re.compile(r"(\[\".*?\"\])|(\[.*?\])")


def _parse(markup: str) -> Iterable[Tuple[Optional[str], Optional[str]]]:
    """Parse markup in to an iterable of pairs of text, tag.
    
    Args:
        markup (str): A string containing console markup
    
    """
    position = 0
    for match in re_tags.finditer(markup):
        escaped_text, tag_text = match.groups()

        start, end = match.span()
        if start > position:
            yield markup[position:start], None
        if tag_text is not None:
            yield None, tag_text
        else:
            yield escaped_text[2:-2], None  # type: ignore

        position = end
    if position < len(markup):
        yield markup[position:], None


def render(markup: str, style: Union[str, Style] = "") -> Text:
    """Render console markup in to a Text instance.

    Args:
        markup (str): A string containing console markup.
    
    Raises:
        MarkupError: If there is a syntax error in the markup.
    
    Returns:
        Text: A test instance.
    """
    text = Text(style=style)
    stylize = text.stylize

    styles: Dict[str, List[int]] = defaultdict(list)
    style_stack: List[str] = []

    for plain_text, tag in _parse(markup):
        if plain_text is not None:
            text.append(plain_text)
        if tag is not None:
            if tag.startswith("[/"):
                style_name = tag[2:-1].strip()
                try:
                    style_position = styles[style_name].pop()
                except (KeyError, IndexError):
                    raise MarkupError(
                        f"closing tag {tag!r} at position {len(text)} doesn't match open tag"
                    )
                style_stack.remove(style_name)
                stylize(style_position, len(text), style_name)
            else:
                style_name = tag[1:-1].strip()
                styles[style_name].append(len(text))
                style_stack.append(style_name)

    text_length = len(text)
    while style_stack:
        style_name = style_stack.pop()
        style_position = styles[style_name].pop()
        text.stylize(style_position, text_length, style_name)

    return text


if __name__ == "__main__":
    text = """Hello [bold red]World[/bold red]!"""

    from .console import Console

    console = Console()
    console.print(render(text))
    console.print(len(render(text)))
    console.print(len(render(text).text))



class ConsoleError(Exception):
    """An error in console operation."""


class StyleError(Exception):
    """An error in styles."""


class StyleSyntaxError(ConsoleError):
    """Style was badly formatted."""


class MissingStyle(StyleError):
    """No such style."""


class StyleStackError(ConsoleError):
    """Style stack is invalid."""


class NotRenderableError(ConsoleError):
    """Object is not renderable."""


class MarkupError(ConsoleError):
    """Markup was badly formatted."""



from typing import Tuple, Union

from .box import Box, SQUARE
from .console import (
    Console,
    ConsoleOptions,
    ConsoleRenderable,
    RenderableType,
    RenderResult,
    RenderWidth,
)
from .style import Style
from .text import Text
from .segment import Segment


class Panel:
    def __init__(
        self,
        renderable: Union[str, ConsoleRenderable],
        box: Box = SQUARE,
        expand: bool = True,
        style: Union[str, Style] = "none",
    ) -> None:
        """A console renderable that draws a border around its contents.
        
        Args:
            renderable (ConsoleRenderable): A console renderable objects.
            box (Box, optional): A Box instance that defines the look of the border.
                Defaults to box.SQUARE.
            expand (bool, optional): If True the panel will stretch to fill the console 
                width, otherwise it will be sized to fit the contents. Defaults to False.
            style (str, optional): The style of the border. Defaults to "none".
        """
        self.renderable = (
            Text.from_markup(renderable) if isinstance(renderable, str) else renderable
        )
        self.box = box
        self.expand = expand
        self.style = style

    def __console__(self, console: Console, options: ConsoleOptions) -> RenderResult:

        style = console.get_style(self.style)
        width = options.max_width

        if self.expand:
            child_width = width - 2
        else:
            print(self.renderable)
            child_width = RenderWidth.get(self.renderable, width - 2).maximum
            print(child_width)

        width = child_width + 2
        child_options = options.update(width=child_width)
        lines = console.render_lines(self.renderable, child_options)

        box = self.box
        line_start = Segment(box.mid_left, style)
        line_end = Segment(f"{box.mid_right}\n", style)
        yield Segment(box.get_top([width - 2]), style)
        for line in lines:
            yield line_start
            yield from line
            yield line_end
        yield Segment(box.get_bottom([width - 2]), style)

    def __console_width__(self, max_width: int) -> RenderWidth:
        if self.expand:
            return RenderWidth(max_width, max_width)
        width = RenderWidth.get(self.renderable, max_width - 2).maximum + 2
        return RenderWidth.get(width, width)


if __name__ == "__main__":
    from .console import Console

    c = Console()

    from .padding import Padding
    from .box import ROUNDED

    p = Panel(
        Panel(
            Padding(Text.from_markup("[bold magenta]Hello World!"), (1, 8)),
            box=ROUNDED,
            expand=False,
        )
    )

    print(p)
    c.print(p)



from typing import NamedTuple

from .segment import Segment


class RenderWidth(NamedTuple):
    """Range of widths for a renderable object."""

    minimum: int
    maximum: int

    @property
    def span(self) -> int:
        """Get difference between maximum and minimum."""
        return self.maximum - self.minimum

    def normalize(self) -> "RenderWidth":
        minimum, maximum = self
        minimum = max(0, minimum)
        return RenderWidth(minimum, max(minimum, maximum))

    def with_maximum(self, width: int) -> "RenderWidth":
        """Get a RenderableWith where the widths are <= width.
        
        Args:
            width (int): Maximum desired width.
        
        Returns:
            RenderableWidth: new RenderableWidth object.
        """
        minimum, maximum = self
        return RenderWidth(min(minimum, width), min(maximum, width))

    @classmethod
    def get(cls, renderable: "RenderableType", max_width: int) -> "RenderWidth":
        """Get desired width for a renderable."""
        if hasattr(renderable, "__console__"):
            get_console_width = getattr(renderable, "__console_width__", None)
            if get_console_width is not None:
                render_width = get_console_width(max_width).with_maximum(max_width)
                return render_width.normalize()
            else:
                return RenderWidth(1, max_width)
        elif isinstance(renderable, Segment):
            text, _style = renderable
            width = min(max_width, len(text))
            return RenderWidth(width, width)
        elif isinstance(renderable, str):
            text = renderable.rstrip()
            return RenderWidth(len(text), len(text))
        else:
            raise errors.NotRenderableError(
                f"Unable to get render width for {renderable!r}; "
                "a str, Segment, or object with __console__ method is required"
            )



from typing import List, Tuple

from .color_triplet import ColorTriplet
from .palette import Palette

_ColorTuple = Tuple[int, int, int]


class Theme:
    """A terminal theme."""

    def __init__(
        self, background: _ColorTuple, foreground: _ColorTuple, ansi: List[_ColorTuple]
    ) -> None:
        self.background_color = ColorTriplet(*background)
        self.foreground_color = ColorTriplet(*foreground)
        self.ansi_colors = Palette(ansi)



import textwrap
from typing import Any, Dict, Union

from pygments.lexers import get_lexer_by_name, guess_lexer_for_filename
from pygments.styles import get_style_by_name
from pygments.token import Token
from pygments.util import ClassNotFound

from .color import Color, parse_rgb_hex, blend_rgb
from .console import Console, ConsoleOptions, RenderResult, Segment, ConsoleRenderable
from .style import Style
from .text import Text
from ._tools import iter_first


class Syntax:
    def __init__(
        self,
        code: str,
        lexer_name: str,
        *,
        style: Union[str, Style] = None,
        theme: str = "emacs",
        dedent: bool = True,
        line_numbers: bool = False,
        start_line: int = 1,
    ) -> None:

        if dedent:
            code = textwrap.dedent(code)
        self.code = code
        self.lexer_name = lexer_name
        self.style = style
        self.theme = theme
        self.dedent = dedent
        self.line_numbers = line_numbers
        self.start_line = start_line

        self._style_cache: Dict[Any, Style] = {}

        try:
            self._pygments_style_class = get_style_by_name(theme)
        except ClassNotFound:
            self._pygments_style_class = get_style_by_name("default")

        self._background_color = self._pygments_style_class.background_color

    @classmethod
    def from_path(
        cls,
        path: str,
        style: Union[str, Style] = None,
        theme: str = "emacs",
        dedent: bool = True,
        line_numbers: bool = False,
        start_line: int = 1,
    ) -> "Syntax":
        """Get a Syntax object for given path."""
        with open(path, "rt") as code_file:
            code = code_file.read()
        try:
            lexer = guess_lexer_for_filename(path, code)
            lexer_name = lexer.name
        except ClassNotFound:
            lexer_name = "default"
        return cls(
            code,
            lexer_name,
            style=style,
            theme=theme,
            dedent=dedent,
            line_numbers=line_numbers,
            start_line=start_line,
        )

    def _get_theme_style(self, token_type) -> Style:
        if token_type in self._style_cache:
            style = self._style_cache[token_type]
        else:
            pygments_style = self._pygments_style_class.style_for_token(token_type)

            color = pygments_style["color"]
            bgcolor = pygments_style["bgcolor"]
            style = Style(
                color="#" + color if color else "#000000",
                bgcolor="#" + bgcolor if bgcolor else self._background_color,
                bold=pygments_style["bold"],
                italic=pygments_style["italic"],
                underline=pygments_style["underline"],
            )
            self._style_cache[token_type] = style

        return style

    def _get_default_style(self) -> Style:
        style = self._get_theme_style(Token.Text)
        style = style + Style(bgcolor=self._pygments_style_class.background_color)
        return style

    def _highlight(self, lexer_name: str) -> Text:
        default_style = self._get_default_style()
        try:
            lexer = get_lexer_by_name(lexer_name)
        except ClassNotFound:
            return Text(self.code, style=default_style)
        text = Text(style=default_style)
        append = text.append
        _get_theme_style = self._get_theme_style
        for token_type, token in lexer.get_tokens(self.code):
            append(token, _get_theme_style(token_type))
        return text

    def _get_line_numbers_color(self) -> Color:
        background_color = parse_rgb_hex(
            self._pygments_style_class.background_color[1:]
        )
        foreground_color = self._get_theme_style(Token.Text)._color
        if foreground_color is None:
            return Color.default()
        # TODO: Handle no full colors here
        assert foreground_color.triplet is not None
        new_color = blend_rgb(background_color, foreground_color.triplet)
        return Color.from_triplet(new_color)

    def __console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        code = self.code
        if self.dedent:
            code = textwrap.dedent(code)
        text = self._highlight(self.lexer_name)
        if not self.line_numbers:
            yield text
            return

        lines = text.split("\n")
        numbers_column_width = len(str(self.start_line + len(lines))) + 2
        render_options = options.update(width=options.max_width - numbers_column_width)

        if self.style is None:
            background_style = Style(
                bgcolor=self._pygments_style_class.background_color
            )
        else:
            if isinstance(self.style, str):
                background_style = console.get_style(self.style)
            else:
                background_style = self.style

        number_style = background_style + self._get_theme_style(Token.Text)
        number_style._color = self._get_line_numbers_color()
        padding = Segment(" " * numbers_column_width, background_style)
        new_line = Segment("\n")
        for line_no, line in enumerate(lines, self.start_line):
            wrapped_lines = console.render_lines(
                line, render_options, style=background_style
            )
            for first, wrapped_line in iter_first(wrapped_lines):
                if first:
                    yield Segment(
                        f" {str(line_no).rjust(numbers_column_width - 2)} ",
                        number_style,
                    )
                else:
                    yield padding
                yield from wrapped_line
                yield new_line


CODE = r'''
    def __console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        """This is a docstring."""
        code = self.code
        if self.dedent:
            code = textwrap.dedent(code)
        text = highlight(code, self.lexer_name, theme=self.theme)
        if not self.line_numbers:
            yield text
            return

        # This is a comment
        lines = text.split("\n")
        numbers_column_width = len(str(len(lines))) + 1
        render_options = options.update(width=options.max_width - numbers_column_width)

        padding = Segment(" " * numbers_column_width)
        new_line = Segment("\n")
        for line_no, line in enumerate(lines, self.start_line):

            wrapped_lines = console.render_lines([line], render_options)
            for first, wrapped_line in iter_first(wrapped_lines):
                if first:
                    yield Segment(f"{line_no}".ljust(numbers_column_width))
                else:
                    yield padding
                yield from wrapped_line
                yield new_line
'''

if __name__ == "__main__":

    syntax = Syntax(CODE, "python", dedent=True, line_numbers=True, start_line=990)

    from time import time

    syntax = Syntax.from_path("./rich/syntax.py", theme="monokai", line_numbers=True)
    console = Console(record=True)
    start = time()
    console.print(syntax)
    elapsed = int((time() - start) * 1000)
    print(f"{elapsed}ms")

    print(Color.downgrade.cache_info())
    print(Color.parse.cache_info())
    print(Color.get_ansi_codes.cache_info())

    print(Style.parse.cache_info())




from datetime import datetime
from typing import Any, Iterable, List, Optional, TYPE_CHECKING

from .text import Text

if TYPE_CHECKING:
    from .console import Console, ConsoleRenderable, RenderableType
    from table import Table


class LogRender:
    def __init__(
        self,
        show_time: bool = True,
        show_path: bool = True,
        time_format: str = "[%x %X] ",
    ) -> None:
        self.show_time = show_time
        self.show_path = show_path
        self.time_format = time_format
        self._last_time: Optional[str] = None

    def __call__(
        self,
        console: "Console",
        renderables: Iterable["ConsoleRenderable"],
        log_time: datetime = None,
        path: str = None,
        line_no: int = None,
    ) -> "Table":
        from .containers import Renderables
        from .table import Table

        output = Table(show_header=False, expand=True, box=None, padding=0)
        if self.show_time:
            output.add_column(style="log.time")
        output.add_column(ratio=1, style="log.message")
        if self.show_path and path:
            output.add_column(style="log.path")
        row: List["RenderableType"] = []
        if self.show_time:
            if log_time is None:
                log_time = datetime.now()
            log_time_display = log_time.strftime(self.time_format)
            if log_time_display == self._last_time:
                row.append(Text(" " * len(log_time_display)))
            else:
                row.append(log_time_display)
                self._last_time = log_time_display
        row.append(Renderables(renderables))
        if self.show_path and path:
            if line_no is None:
                row.append(Text(path))
            else:
                row.append(Text(f"{path}:{line_no}"))
        output.add_row(*row)
        return output


if __name__ == "__main__":  # pragma: no cover
    from .console import Console

    console = Console()
    print(console)
    logger = Logger(console)

    from .markdown import Markdown
    from .syntax import Syntax

    s = Syntax(
        '''\
@classmethod
def get(cls, renderable: RenderableType, max_width: int) -> RenderWidth:
    """Get desired width for a renderable."""
    if hasattr(renderable, "__console__"):
        get_console_width = getattr(renderable, "__console_width__", None)
        if get_console_width is not None:
            render_width = get_console_width(max_width).with_maximum(max_width)
            return render_width.normalize()
        else:
            return RenderWidth(1, max_width)
    elif isinstance(renderable, Segment):
        text, _style = renderable
        width = min(max_width, len(text))
        return RenderWidth(width, width)
    elif isinstance(renderable, str):
        text = renderable.rstrip()
        return RenderWidth(len(text), len(text))
    else:
        raise errors.NotRenderableError(
            f"Unable to get render width for {renderable!r}; "
            "a str, Segment, or object with __console__ method is required"
        )
        ''',
        "python",
        theme="monokai",
    )

    logger(
        "Hello", path="foo.py", line_no=20,
    )
    logger(
        "World!", path="foo.py", line_no=20,
    )



from typing import Iterable, List, Tuple, TypeVar

T = TypeVar("T")


def iter_first(values: Iterable[T]) -> Iterable[Tuple[bool, T]]:
    """Iterate and generate a tuple with a flag for first value."""
    iter_values = iter(values)
    try:
        value = next(iter_values)
    except StopIteration:
        return
    yield True, value
    for value in iter_values:
        yield False, value


def iter_last(values: Iterable[T]) -> Iterable[Tuple[bool, T]]:
    """Iterate and generate a tuple with a flag for last value."""
    iter_values = iter(values)
    try:
        previous_value = next(iter_values)
    except StopIteration:
        return
    for value in iter_values:
        yield False, previous_value
        previous_value = value
    yield True, previous_value


def iter_first_last(values: Iterable[T]) -> Iterable[Tuple[bool, bool, T]]:
    """Iterate and generate a tuple with a flag for first and last value."""
    iter_values = iter(values)
    try:
        previous_value = next(iter_values)
    except StopIteration:
        return
    first = True
    for value in iter_values:
        yield first, False, previous_value
        first = False
        previous_value = value
    yield first, True, previous_value


def ratio_divide(
    total: int, ratios: List[int], minimums: List[int] = None
) -> List[int]:
    """Divide an integer total in to parts based on ratios.
    
    Args:
        total (int): The total to divide.
        ratios (List[int]): A list of integer rations.
        minimums (List[int]): List of minimum values for each slot. 
    
    Returns:
        List[int]: A list of integers garanteed to sum to total.
    """
    total_ratio = sum(ratios)
    assert total_ratio > 0, "Sum of ratios must be > 0"

    total_remaining = total
    distributed_total: List[int] = []
    append = distributed_total.append
    if minimums is None:
        _minimums = [0] * len(ratios)
    else:
        _minimums = minimums
    for ratio, minimum in zip(ratios, _minimums):
        if total_ratio > 0:
            distributed = max(minimum, round(ratio * total_remaining / total_ratio))
        else:
            distributed = total_remaining
        append(distributed)
        total_ratio -= ratio
        total_remaining -= distributed
    return distributed_total


if __name__ == "__main__":  # pragma: no coverage

    n = ["foo", "bar", "egg", "baz"]
    for first, last, t in iter_first_last(n):
        print(first, last, t)

    print(ratio_divide(7, [1]))
    print(ratio_divide(10, [1, 2, 1]))
    print(ratio_divide(3, [1, 2, 1]))
    print(ratio_divide(20, [1, 2]))
    print(ratio_divide(7, [1, 0, 0]))



from typing import cast, Tuple, TYPE_CHECKING, Union

if TYPE_CHECKING:
    from .console import (
        Console,
        ConsoleOptions,
        RenderableType,
        RenderResult,
    )
from ._render_width import RenderWidth
from .style import Style
from .text import Text
from .segment import Segment


PaddingDimensions = Union[int, Tuple[int], Tuple[int, int], Tuple[int, int, int, int]]


class Padding:
    """Draw space around a console renderable."""

    def __init__(
        self,
        renderable: "RenderableType",
        pad: "PaddingDimensions",
        *,
        style: Union[str, Style] = "none",
    ):
        self.renderable = renderable
        self.top, self.right, self.bottom, self.left = self.unpack(pad)
        self.style = style

    @staticmethod
    def unpack(pad: "PaddingDimensions") -> Tuple[int, int, int, int]:
        """Unpack padding specified in CSS style."""
        if isinstance(pad, int):
            return (pad, pad, pad, pad)
        if len(pad) == 1:
            _pad = pad[0]
            return (_pad, _pad, _pad, _pad)
        if len(pad) == 2:
            pad_top, pad_right = cast(Tuple[int, int], pad)
            return (pad_top, pad_right, pad_top, pad_right)
        if len(pad) == 3:
            raise ValueError(
                f"1, 2 or 4 integers required for padding; {len(pad)} given"
            )
        if len(pad) == 4:
            top, right, bottom, left = cast(Tuple[int, int, int, int], pad)
            return (top, right, bottom, left)
        raise ValueError(f"1, 2 or 4 integers required for padding; {len(pad)} given")

    def __repr__(self) -> str:
        return f"Padding({self.renderable!r}, ({self.top},{self.right},{self.bottom},{self.left}))"

    def __console__(
        self, console: "Console", options: "ConsoleOptions"
    ) -> "RenderResult":

        style = console.get_style(self.style)
        width = options.max_width
        child_options = options.update(width=width - self.left - self.right)
        lines = console.render_lines(self.renderable, child_options, style=style)
        lines = Segment.set_shape(lines, child_options.max_width, style=style)

        blank_line = Segment(" " * width + "\n", style)
        top = [blank_line] * self.top
        bottom = [blank_line] * self.top
        left = Segment(" " * self.left, style) if self.left else None
        right = Segment(" " * self.right, style) if self.right else None
        new_line = Segment.line()
        yield from top
        for line in lines:
            if left is not None:
                yield left
            yield from line
            if right is not None:
                yield right
            yield new_line
        yield from bottom

    def __console_width__(self, max_width: int) -> "RenderWidth":
        extra_width = self.left + self.right
        min_width, max_width = RenderWidth.get(
            self.renderable, max(1, max_width - extra_width)
        )
        render_width = RenderWidth(min_width + extra_width, max_width + extra_width)
        return render_width




from collections.abc import Sequence, Mapping
from typing import Any, TYPE_CHECKING


from . import box
from .repr import console_repr
from .table import Column, Table
from .text import Text


if TYPE_CHECKING:
    from .console import ConsoleRenderable


def inspect(object: Any) -> ConsoleRenderable:
    if isinstance(object, str):
        return console_repr(object)
    if isinstance(object, Mapping):
        return inspect_mapping(object)
    if isinstance(object, Sequence):
        return inspect_sequence(object)
    return console_repr(object)


def inspect_sequence(sequence: Sequence) -> Table:
    table = Table(
        Column("key", style="inspect.key", justify="right"),
        Column("value", style="inspect.value"),
        show_header=False,
        box=None,
        padding=0,
    )
    for key, value in enumerate(sequence):
        table.add_row(console_repr(key) + Text(": ", "not italic"), console_repr(value))
    return table


def inspect_mapping(mapping: Mapping) -> Table:
    table = Table(
        Column("key", style="inspect.key"),
        Column("value", style="inspect.value"),
        show_header=False,
        box=None,
        padding=0,
    )
    for key, value in mapping.items():
        table.add_row(console_repr(key) + Text(": ", "not italic"), console_repr(value))
    return table


if __name__ == "__main__":
    from .console import Console

    console = Console()
    mapping = {
        "version": "1.1",
        "method": "confirmFruitPurchase",
        "params": [["apple", "orange", "mangoes"], 1.123],
        "id": "194521489",
    }
    console.print(inspect(mapping))
    console.print(inspect(["Hello", "World", "!", 5] * 6))

    from .syntax import Syntax

    syntax = Syntax(repr(mapping), "python", theme="monokai")
    console.print(syntax)
    console.print(Text("50", style="italic bold") + Text(":"))

    from pprint import pprint

    pprint(mapping)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rich']

package_data = \
{'': ['*']}

install_requires = \
['pprintpp>=0.4.0,<0.5.0', 'typing-extensions>=3.7.4,<4.0.0']

extras_require = \
{':python_version >= "3.6" and python_version < "3.7"': ['dataclasses>=0.7,<0.8']}

setup_kwargs = {
    'name': 'rich',
    'version': '0.2.1',
    'description': 'Render rich text, tables, syntax highlighting, markdown and more to the terminal',
    'long_description': '**Note:** This library is currently work in progress. Documentation and tests are coming.\n\n# Rich\n\nRich is a Python library for _rich_ text and high level formatting in the terminal.\n\nThe Rich API make it easy to add colored text (up to 16.7million colors) and styles (bold, italic, underline etc.) to your script or application. Rich can also render pretty tables, markdown and source code with syntax highlighting.\n\n## Installing\n\nInstall with `pip` or your favorite PyPi package manager.\n\n```\npip install rich\n```\n\n## Console Printing\n\nThe first step to using the rich console is to import and construct the `Console` object.\n\n```python\nfrom rich.console import Console\n\nconsole = Console()\n```\n\nMost applications will require one `Console` instance. The easiest way to manage your console instance would be to construct an instance at the module level and import it where needed.\n\nThe Console object has a `print` method which has an intentionally similar interface to the builtin `print` function. Here\'s an example of use:\n\n```\nconsole.print("Hello", "World!")\n```\n\nAs you might expect, this will print `"Hello World!"` to the terminal. Note that unlike the `print` function, Rich will word-wrap your test to fit within the terminal width.\n\nThere are a few ways of adding color and style to your output. You can set a style for the entire output by adding a `style` keyword argument. Here\'s an example:\n\n```\nconsole.print("Hello", "World!", style="bold red")\n```\n\nThe output will be something like the following:\n\n![Hello World](./imgs/hello_world.png)\n\nThat\'s fine for styling a line of text at a time. For more finely grained styling, Rich renders a special markup which is similar in syntax to [bbcode](https://en.wikipedia.org/wiki/BBCode). Here\'s an example:\n\n```python\nconsole.print("Where there is a [bold cyan]Will[/bold cyan] there [u]is[/u] a [i]way[/i].")\n```\n\n![Console Markup](./imgs/where_there_is_a_will.png)\n\n### Defining Styles\n\nRich expects styles to be specified with a _style definition_ string, which is a intuitive syntax that reads almost like English.\n\nThere are a few ways of specifying colors:\n\n- The word `"default"` to select the default foreground or background color.\n- The name of the color (one of 256 possible constants) e.g. `"magenta"`. To see the full range of colors run `python -m rich.color` from the console.\n- A number between 0 and 255 (inclusive) which corresponds to one of the 256 possible constants above.\n- A CSS hex style color, e.g. `#ff0000` or `#d75faf`\n- A CSS rbg style color, e.g. `rgb(215,95,175)`\n\nBy itself, a color will set the _foreground_ color. To set a _background_ color, precede the color with the word `"on"`. For example `"red on white"`.\n\nNote that the CSS hex and RGB style of color lets you chose one of 16.7 million colors, but some terminals (notably OSX terminal) only support 256 colors. If Rich detects that only 256 colors are supported it will pick the closest color available. In practice this means that you may not get exactly the color you ask for, but it is generally _close enough_.\n\nTo set a style or attribute add one or more of the following words:\n\n- `"bold"` for bold text.\n- `"dim"` for dim text.\n- `"italic"` for italic text.\n- `"underline"` for underlined text.\n- `"blink"` for text that blinks.\n- `"reverse"` to swap foreground and background text.\n- `"conceal"` for concealed text (not supported on most terminals).\n- `"strike"` for text with a line through it (not supported on all terminals).\n\nStyle attributes and colors may appear in any order, i.e. `"bold magenta on yellow"` has the same effect as `"on yellow magenta bold"`. The latter may be preferred by Yoda.\n\n## Console Logging\n\nThe Console object has a `log()` method which has a similar interface to `print()`, but also renders a column for the current time and the file and line which made the call. By default, Rich will do syntax highlighting for Python structures and for repr strings. If you log a collection (i.e. a dict or a list) Rich will pretty print it so that it fits in the available space. Here\'s an example of some of these features.\n\n```python\nfrom rich.console import Console\nconsole = Console()\n\ntest_data = [\n    {"jsonrpc": "2.0", "method": "sum", "params": [None, 1, 2, 4, False, True], "id": "1",},\n    {"jsonrpc": "2.0", "method": "notify_hello", "params": [7]},\n    {"jsonrpc": "2.0", "method": "subtract", "params": [42, 23], "id": "2"},\n]\n\ndef test_log():\n    enabled = False\n    context = {\n        "foo": "bar",\n    }\n    movies = ["Deadpool", "Rise of the Skywalker"]\n    console.log("Hello from", console, "!")\n    console.log(test_data, log_locals=True)\n\n\ntest_log()\n```\n\nThe above produces the following output:\n\n![Log](./imgs/log.png)\n\nNote the `log_locals` argument, which outputs a table containing the local variables where the log method was called.\n\nThe log method could be used for logging to the terminal for long running applications such as servers, but is also a very nice debugging aid.\n\n## Emoji\n\nTo insert an emoji in to console output place the name between two colons. Here\'s an example:\n\n```python\n>>> console.print(":smiley: :vampire: :pile_of_poo: :thumbs_up: :raccoon:")\n😃 🧛 💩 👍 🦝\n```\n\nPlease use this feature wisely.\n\n## Markdown\n\nRich can render markdown and does a reasonable job of translating the formatting to the terminal.\n\nTo render markdown import the `Markdown` class and construct it with a string containing markdown code. Then print it to the console. Here\'s an example:\n\n```python\nfrom rich.console import Console\nfrom rich.markdown import Markdown\n\nconsole = Console()\nwith open("README.md") as readme:\n    markdown = Markdown(readme.read())\nconsole.print(markdown)\n```\n\nThis will produce output something like the following:\n\n![markdown](./imgs/markdown.png)\n\n## Syntax Highlighting\n\nRich uses the [pygments](https://pygments.org/) library to implement syntax highlighting. Usage is similar to rendering markdown; construct a `Syntax` object and print it to the console. Here\'s an example:\n\n```python\nfrom rich.console import Console\nfrom rich.syntax import Syntax\n\nmy_code = \'\'\'\ndef iter_first_last(values: Iterable[T]) -> Iterable[Tuple[bool, bool, T]]:\n    """Iterate and generate a tuple with a flag for first and last value."""\n    iter_values = iter(values)\n    try:\n        previous_value = next(iter_values)\n    except StopIteration:\n        return\n    first = True\n    for value in iter_values:\n        yield first, False, previous_value\n        first = False\n        previous_value = value\n    yield first, True, previous_value\n\'\'\'\nsyntax = Syntax(my_code, "python", theme="monokai", line_numbers=True)\nconsole = Console()\nconsole.print(syntax)\n```\n\nThis will produce the following output:\n\n![syntax](./imgs/syntax.png)\n\n## Tables\n\nRich can render flexible tables with unicode box characters. There is a large variety of formatting options for borders, styles, cell alignment etc. Here\'s a simple example:\n\n```python\nfrom rich.console import Console\nfrom rich.table import Column, Table\n\nconsole = Console()\n\ntable = Table(show_header=True, header_style="bold magenta")\ntable.add_column("Date", style="dim", width=12)\ntable.add_column("Title")\ntable.add_column("Production Budget", justify="right")\ntable.add_column("Box Office", justify="right")\ntable.add_row(\n    "Dev 20, 2019", "Star Wars: The Rise of Skywalker", "$275,000,0000", "$375,126,118"\n)\ntable.add_row(\n    "May 25, 2018",\n    "[red]Solo[/red]: A Star Wars Story",\n    "$275,000,0000",\n    "$393,151,347",\n)\ntable.add_row(\n    "Dec 15, 2017",\n    "Star Wars Ep. VIII: The Last Jedi",\n    "$262,000,000",\n    "[bold]$1,332,539,889[/bold]",\n)\n\nconsole.print(table)\n```\n\nThis produces the following output:\n\n![table](./imgs/table.png)\n\nNote that console markup is rendered in the same was as `print()` and `log()`. In fact, anything that is renderable by Rich may be included in the headers / rows (even other tables).\n\nThe `Table` class is smart enough to resize columns to fit the available width of the terminal, wrapping text as required. Here\'s the same example, with the terminal made smaller than the table above:\n\n![table2](./imgs/table2.png)\n',
    'author': 'Will McGugan',
    'author_email': 'willmcgugan@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://www.willmcgugan.com/search/?s=rich',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from typing import Iterable, List
from typing_extensions import Literal

from ._tools import iter_last


class Box:
    """Defines characters to render boxes.
    
    ┌─┬┐ top
    │ ││ head
    ├─┼┤ head_row
    │ ││ mid
    ├─┼┤ row
    ├─┼┤ foot_row
    │ ││ foot
    └─┴┘ bottom

    
    """

    def __init__(self, box: str) -> None:
        self._box = box
        (line1, line2, line3, line4, line5, line6, line7, line8,) = box.splitlines()
        # top
        self.top_left, self.top, self.top_divider, self.top_right = line1
        # head
        self.head_left, _, self.head_vertical, self.head_right = line2
        # head_row
        (
            self.head_row_left,
            self.head_row_horizontal,
            self.head_row_cross,
            self.head_row_right,
        ) = line3

        # mid
        self.mid_left, _, self.mid_vertical, self.mid_right = line4
        # row
        self.row_left, self.row_horizontal, self.row_cross, self.row_right = line5
        # foot_row
        (
            self.foot_row_left,
            self.foot_row_horizontal,
            self.foot_row_cross,
            self.foot_row_right,
        ) = line6
        # foot
        self.foot_left, _, self.foot_vertical, self.foot_right = line7
        # bottom
        self.bottom_left, self.bottom, self.bottom_divider, self.bottom_right = line8

    def __repr__(self) -> str:
        border_str = str(self).replace("\n", r"\n")
        return f'Border("{border_str}")'

    def __str__(self) -> str:
        return self._box

    def get_top(self, widths: Iterable[int]) -> str:
        """Get the top of a simple box.
        
        Args:
            widths (List[int]): Widths of columns.
        
        Returns:
            str: A string of box characters.
        """

        parts: List[str] = []
        append = parts.append
        append(self.top_left)
        for last, width in iter_last(widths):
            append(self.top * width)
            if not last:
                append(self.top_divider)
        append(self.top_right)
        append("\n")
        return "".join(parts)

    def get_row(
        self,
        widths: Iterable[int],
        level: Literal["head", "row", "foot"] = "row",
        edge: bool = True,
    ) -> str:
        """Get the top of a simple box.
        
        Args:
            width (List[int]): Widths of columns.
        
        Returns:
            str: A string of box characters.
        """
        if level == "head":
            left = self.head_row_left
            horizontal = self.head_row_horizontal
            cross = self.head_row_cross
            right = self.head_row_right
        elif level == "row":
            left = self.row_left
            horizontal = self.row_horizontal
            cross = self.row_cross
            right = self.row_right
        elif level == "foot":
            left = self.foot_row_left
            horizontal = self.foot_row_horizontal
            cross = self.foot_row_cross
            right = self.foot_row_right
        else:
            raise ValueError("level must be 'head', 'row' or 'foot'")

        parts: List[str] = []
        append = parts.append
        if edge:
            append(left)
        for last, width in iter_last(widths):
            append(horizontal * width)
            if not last:
                append(cross)
        if edge:
            append(right)
        append("\n")
        return "".join(parts)

    def get_bottom(self, widths: Iterable[int]) -> str:
        """Get the top of a simple box.
        
        Args:
            widths (List[int]): Widths of columns.
        
        Returns:
            str: A string of box characters.
        """

        parts: List[str] = []
        append = parts.append
        append(self.bottom_left)
        for last, width in iter_last(widths):
            append(self.bottom * width)
            if not last:
                append(self.bottom_divider)
        append(self.bottom_right)
        append("\n")
        return "".join(parts)


ASCII = Box(
    """\
+--+
| ||
|-+|
| ||
|-+|
|-+|
| ||
+--+
"""
)

SQUARE = Box(
    """\
┌─┬┐
│ ││
├─┼┤
│ ││
├─┼┤
├─┼┤
│ ││
└─┴┘
"""
)


MINIMAL = Box(
    """\
    
  │ 
 ─┼ 
  │ 
 ─┼ 
 ─┼ 
  │ 
    
"""
)


MINIMAL_HEAVY_HEAD = Box(
    """\
    
  │ 
 ━┿ 
  │ 
 ─┼ 
 ─┼ 
  │ 
    
"""
)

MINIMAL_DOUBLE_HEAD = Box(
    """\
    
  │ 
 ═╪ 
  │ 
 ─┼ 
 ─┼ 
  │ 
    
"""
)


SIMPLE = Box(
    """\
    
    
────
    
    
────
    
    
"""
)


SIMPLE_HEAVY = Box(
    """\
    
    
╺━━╸
    
    
╺━━╸
    
    
"""
)


HORIZONTALS = Box(
    """\
────
    
────
    
────
────
    
────
"""
)

ROUNDED = Box(
    """\
╭─┬╮
│ ││
├─┼┤
│ ││
├─┼┤
├─┼┤
│ ││
╰─┴╯
"""
)

HEAVY = Box(
    """\
┏━┳┓
┃ ┃┃
┣━╋┫
┃ ┃┃
┣━╋┫
┣━╋┫
┃ ┃┃
┗━┻┛
"""
)

HEAVY_EDGE = Box(
    """\
┏━┯┓
┃ │┃
┠─┼┨
┃ │┃
┠─┼┨
┠─┼┨
┃ │┃
┗━┷┛
"""
)

HEAVY_HEAD = Box(
    """\
┏━┳┓
┃ ┃┃
┡━╇┩
│ ││
├─┼┤
├─┼┤
│ ││
└─┴┘
"""
)

DOUBLE = Box(
    """\
╔═╦╗
║ ║║
╠═╬╣
║ ║║
╠═╬╣
╠═╬╣
║ ║║
╚═╩╝
"""
)

DOUBLE_EDGE = Box(
    """\
╔═╤╗
║ │║
╟─┼╢
║ │║
╟─┼╢
╟─┼╢
║ │║
╚═╧╝
"""
)


if __name__ == "__main__":  # pragma: no cover
    print("ASCII")
    print(ASCII)

    print("SQUARE")
    print(SQUARE)

    print("HORIZONTALS")
    print(HORIZONTALS)

    print("ROUNDED")
    print(ROUNDED)

    print("HEAVY")
    print(HEAVY)

    print("HEAVY_EDGE")
    print(HEAVY_EDGE)

    print("HEAVY_HEAD")
    print(HEAVY_HEAD)

    print("DOUBLE")
    print(DOUBLE)

    print("DOUBLE_EDGE")
    print(DOUBLE_EDGE)



from dataclasses import dataclass, field
from itertools import chain
from typing import (
    Iterable,
    List,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    TYPE_CHECKING,
    Union,
)
from typing_extensions import Literal

from . import box

if TYPE_CHECKING:
    from .console import (
        Console,
        ConsoleOptions,
        ConsoleRenderable,
        JustifyValues,
        RenderableType,
        RenderResult,
    )
from . import errors
from .padding import Padding, PaddingDimensions
from .segment import Segment
from .style import Style
from .text import Text
from ._render_width import RenderWidth
from ._tools import iter_first_last, iter_first, iter_last, ratio_divide


T = TypeVar("T")


def _pick_first(values: Iterable[Optional[T]], final: T) -> T:
    """Pick first non-None value."""
    for value in values:
        if value is not None:
            return value
    return final


@dataclass
class Column:
    """Defines a column in a table."""

    header: Union[str, "ConsoleRenderable"] = ""
    footer: Union[str, "ConsoleRenderable"] = ""
    header_style: Union[str, Style] = "table.header"
    footer_style: Union[str, Style] = "table.footer"
    style: Union[str, Style] = "none"
    justify: "JustifyValues" = "left"
    width: Optional[int] = None
    ratio: Optional[int] = None
    _cells: List["ConsoleRenderable"] = field(default_factory=list)

    @property
    def cells(self) -> Iterable["ConsoleRenderable"]:
        """Get all cells in the column, not including header."""
        yield from self._cells

    @property
    def flexible(self) -> bool:
        """Check if this column is flexible."""
        return self.ratio is not None

    @property
    def header_renderable(self) -> "ConsoleRenderable":
        return (
            Text.from_markup(self.header, style=self.header_style or "")
            if isinstance(self.header, str)
            else self.header
        )

    @property
    def footer_renderable(self) -> "ConsoleRenderable":
        return (
            Text.from_markup(self.footer, style=self.footer_style or "")
            if isinstance(self.footer, str)
            else self.footer
        )


class _Cell(NamedTuple):

    style: Union[str, Style]
    renderable: "ConsoleRenderable"


class Table:
    columns: List[Column]

    def __init__(
        self,
        *headers: Union[Column, str],
        title: str = None,
        footer: str = None,
        width: int = None,
        box: Optional[box.Box] = box.DOUBLE_EDGE,
        padding: PaddingDimensions = (0, 1),
        pad_edge: bool = True,
        expand: bool = False,
        show_header: bool = True,
        show_footer: bool = False,
        show_edge: bool = True,
        style: Union[str, Style] = "none",
        header_style: Union[str, Style] = None,
        footer_style: Union[str, Style] = None,
        border_style: Union[str, Style] = None,
        title_style: Union[str, Style] = None,
    ) -> None:
        self.columns = [
            (Column(header) if isinstance(header, str) else header)
            for header in headers
        ]
        self.title = title
        self.width = width
        self.box = box
        self._padding = Padding.unpack(padding)
        self.pad_edge = pad_edge
        self.expand = expand
        self.show_header = show_header
        self.show_footer = show_footer
        self.show_edge = show_edge
        self.style = style
        self.header_style = header_style
        self.footer_style = footer_style
        self.border_style = border_style
        self.title_style = title_style
        self._row_count = 0

    @property
    def padding(self) -> Tuple[int, int, int, int]:
        return self._padding

    @padding.setter
    def padding(self, padding: PaddingDimensions) -> "Table":
        self._padding = Padding.unpack(padding)
        return self

    def add_column(
        self,
        header: Union[str, "ConsoleRenderable"] = "",
        footer: Union[str, "ConsoleRenderable"] = "",
        header_style: Union[str, Style] = None,
        footer_style: Union[str, Style] = None,
        style: Union[str, Style] = None,
        justify: "JustifyValues" = "left",
        width: int = None,
        ratio: int = None,
    ):
        """Add a column to the table.
        
        Args:
            header (Union[str, ConsoleRenderable], optional): Text or renderable for the header.
                Defaults to "".
            footer (Union[str, ConsoleRenderable], optional): Text or renderable for the footer.
                Defaults to "".
            header_style (Union[str, Style], optional): Style for the header. Defaults to "none".
            footer_style (Union[str, Style], optional): Style for the header. Defaults to "none".
            style (Union[str, Style], optional): Style for the column cells. Defaults to "none".
            justify (JustifyValues, optional): Alignment for cells. Defaults to "left".
            width (int, optional): A minimum width in characters. Defaults to None.
            ratio (int, optional): Flexible ratio for the column. Defaults to None.
        """
        column = Column(
            header=header,
            footer=footer,
            header_style=_pick_first((header_style, self.header_style), "table.header"),
            footer_style=_pick_first((footer_style, self.footer_style), "table.footer"),
            style=_pick_first((style, self.style), "table.cell"),
            justify=justify,
            width=width,
            ratio=ratio,
        )
        self.columns.append(column)

    def add_row(self, *renderables: Optional[Union[str, "ConsoleRenderable"]]) -> None:
        from .console import ConsoleRenderable

        def add_cell(column: Column, renderable: ConsoleRenderable):
            column._cells.append(renderable)

        cell_renderables: List[Optional[Union[str, ConsoleRenderable]]] = list(
            renderables
        )

        columns = self.columns
        if len(cell_renderables) < len(columns):
            cell_renderables = [
                *cell_renderables,
                *[None] * (len(columns) - len(cell_renderables)),
            ]
        for index, renderable in enumerate(cell_renderables):
            if index == len(columns):
                column = Column()
                for _ in range(self._row_count):
                    add_cell(column, Text(""))
                self.columns.append(column)
            else:
                column = columns[index]
            if isinstance(renderable, ConsoleRenderable):
                add_cell(column, renderable)
            elif renderable is None:
                add_cell(column, Text(""))
            elif isinstance(renderable, str):
                add_cell(column, Text.from_markup(renderable))
            else:
                raise errors.NotRenderableError(
                    f"unable to render {renderable!r}; str or object with a __console__ method is required"
                )
        self._row_count += 1

    def __console__(
        self, console: "Console", options: "ConsoleOptions"
    ) -> "RenderResult":

        max_width = options.max_width
        if self.width is not None:
            max_width = min(self.width, max_width)

        if self.box:
            max_width -= len(self.columns)
            if self.show_edge:
                max_width -= 2
        widths = self._calculate_column_widths(max_width)
        table_width = sum(widths) + len(self.columns) + (2 if self.box else 0)
        if self.title:
            title_text = Text.from_markup(self.title, style=self.title_style or "")
            wrapped_title = title_text.wrap(table_width, "center")
            yield wrapped_title
        yield from self._render(console, options, widths)

    def _calculate_column_widths(self, max_width: int) -> List[int]:
        """Calculate the widths of each column."""
        columns = self.columns
        width_ranges = [
            self._measure_column(column_index, column, max_width)
            for column_index, column in enumerate(columns)
        ]

        widths = [_range.maximum or 1 for _range in width_ranges]
        padding_width = self.padding[1] + self.padding[3]
        if self.expand:
            ratios = [col.ratio or 0 for col in columns if col.flexible]
            if any(ratios):
                fixed_widths = [_range.maximum for _range in width_ranges]
                fixed_widths = [
                    _range.minimum if column.flexible else _range.maximum
                    for _range, column in zip(width_ranges, columns)
                ]
                flex_minimum = [
                    (column.width or 0) + padding_width
                    for column in columns
                    if column.flexible
                ]
                flexible_width = max_width - sum(fixed_widths)

                flex_widths = ratio_divide(flexible_width, ratios, flex_minimum)
                iter_flex_widths = iter(flex_widths)
                for index, column in enumerate(columns):
                    if column.flexible:
                        widths[index] = fixed_widths[index] + next(iter_flex_widths)
        table_width = sum(widths)

        if table_width > max_width:
            flex_widths = [_range.span for _range in width_ranges]
            if not any(flex_widths):
                flex_widths = [1] * len(flex_widths)
            excess_width = table_width - max_width
            widths = [
                width - excess_width
                for width, excess_width in zip(
                    widths, ratio_divide(excess_width, flex_widths)
                )
            ]
        elif table_width < max_width and self.expand:
            pad_widths = ratio_divide(max_width - table_width, widths)
            widths = [_width + pad for _width, pad in zip(widths, pad_widths)]

        return widths

    def _get_cells(self, column_index: int, column: Column) -> Iterable[_Cell]:
        """Get all the cells with padding and optional header."""

        padding = self.padding
        any_padding = any(padding)

        first = column_index == 0
        last = column_index == len(self.columns) - 1

        def add_padding(renderable: "ConsoleRenderable") -> "ConsoleRenderable":
            if not any_padding:
                return renderable
            top, right, bottom, left = padding
            if not self.pad_edge:
                if first:
                    left = 0
                if last:
                    right = 0
            return Padding(renderable, (top, right, bottom, left))

        if self.show_header:
            yield _Cell(column.header_style, add_padding(column.header_renderable))
        for cell in column.cells:
            yield _Cell(column.style, add_padding(cell))
        if self.show_footer:
            yield _Cell(column.footer_style, add_padding(column.footer_renderable))

    def _measure_column(
        self, column_index: int, column: Column, max_width: int
    ) -> RenderWidth:
        """Get the minimum and maximum width of the column."""
        padding_width = self.padding[1] + self.padding[3]
        if column.width is not None:
            # Fixed width column
            return RenderWidth(
                column.width + padding_width, column.width + padding_width
            )
        # Flexible column, we need to measure contents
        min_widths: List[int] = []
        max_widths: List[int] = []
        append_min = min_widths.append
        append_max = max_widths.append
        get_render_width = RenderWidth.get
        for cell in self._get_cells(column_index, column):
            _min, _max = get_render_width(cell.renderable, max_width)
            append_min(_min)
            append_max(_max)
        return RenderWidth(
            max(min_widths) if min_widths else 1,
            max(max_widths) if max_widths else max_width,
        )

    def _render(
        self, console: "Console", options: "ConsoleOptions", widths: List[int]
    ) -> "RenderResult":
        table_style = console.get_style(self.style or "")

        border_style = table_style + console.get_style(self.border_style or "")
        rows: Iterable[Tuple[_Cell, ...]] = zip(
            *(
                self._get_cells(column_index, column)
                for column_index, column in enumerate(self.columns)
            )
        )
        box = self.box
        new_line = Segment.line()

        columns = self.columns
        show_header = self.show_header
        show_footer = self.show_footer
        show_edge = self.show_edge

        if box and show_edge:
            yield Segment(box.get_top(widths), border_style)

        for first, last, row in iter_first_last(rows):
            max_height = 1
            cells: List[List[List[Segment]]] = []
            for width, cell, column in zip(widths, row, columns):
                render_options = options.update(width=width, justify=column.justify)
                cell_style = table_style + console.get_style(cell.style)
                lines = console.render_lines(
                    cell.renderable, render_options, style=cell_style
                )
                max_height = max(max_height, len(lines))
                cells.append(lines)

            cells[:] = [
                Segment.set_shape(_cell, width, max_height, style=table_style)
                for width, _cell in zip(widths, cells)
            ]

            if box:
                if last and show_footer:
                    yield Segment(
                        box.get_row(widths, "foot", edge=show_edge), border_style
                    )
                if first:
                    left = Segment(box.head_left, border_style)
                    right = Segment(box.head_right, border_style)
                    divider = Segment(box.head_vertical, border_style)
                elif last:
                    left = Segment(box.foot_left, border_style)
                    right = Segment(box.foot_right, border_style)
                    divider = Segment(box.foot_vertical, border_style)
                else:
                    left = Segment(box.mid_left, border_style)
                    right = Segment(box.mid_right, border_style)
                    divider = Segment(box.mid_vertical, border_style)

                for line_no in range(max_height):
                    if show_edge:
                        yield left
                    for last, rendered_cell in iter_last(cells):
                        yield from rendered_cell[line_no]
                        if not last:
                            yield divider
                    if show_edge:
                        yield right
                    yield new_line
            else:
                for line_no in range(max_height):
                    for rendered_cell in cells:
                        yield from rendered_cell[line_no]
                    yield new_line
            if box and first and show_header:
                yield Segment(box.get_row(widths, "head", edge=show_edge), border_style)

        if box and show_edge:
            yield Segment(box.get_bottom(widths), border_style)


if __name__ == "__main__":  # pragma: no cover

    from .console import Console
    from . import box

    c = Console()
    table = Table(
        Column(
            "Foo", footer=Text("Total", justify="right"), footer_style="bold", ratio=1
        ),
        Column("Bar", style="red", footer="123", ratio=1),
        box=box.SIMPLE,
        expand=True,
        show_footer=True,
        show_edge=True,
    )
    # table.columns[0].width = 50
    # table.columns[1].ratio = 1

    table.add_row("Hello, World! " * 8, "cake" * 10)
    from .markdown import Markdown

    table.add_row(Markdown("# This is *Markdown*!"), "More text", "Hello WOrld")
    table.columns[0].justify = "center"
    table.columns[1].justify = "right"

    c.print(table)



from collections import ChainMap
from collections.abc import Mapping, Sequence
from contextlib import contextmanager
from dataclasses import dataclass, replace
from enum import Enum
import inspect
from itertools import chain
import os
from operator import itemgetter
import re
import shutil
import sys
from typing import (
    Any,
    Callable,
    Dict,
    IO,
    Iterable,
    List,
    Optional,
    NamedTuple,
    overload,
    Tuple,
    TYPE_CHECKING,
    Union,
)
from typing_extensions import Protocol, runtime_checkable, Literal


from ._emoji_replace import _emoji_replace

from . import markup
from ._render_width import RenderWidth
from ._log_render import LogRender
from .default_styles import DEFAULT_STYLES
from . import errors
from .color import ColorSystem
from .highlighter import ReprHighlighter
from .pretty import Pretty
from .style import Style
from .tabulate import tabulate_mapping
from . import highlighter
from . import themes
from .pretty import Pretty
from .theme import Theme
from .segment import Segment

if TYPE_CHECKING:
    from .text import Text

    HighlighterType = Callable[[Union[str, Text]], Text]

JustifyValues = Optional[Literal["left", "center", "right", "full"]]


CONSOLE_HTML_FORMAT = """\
<!DOCTYPE html>
<head>
<style>
{stylesheet}
body {{
    color: {foreground};
    background-color: {background};
}}
</style>
</head>
<html>
<body>
    <code>
        <pre style="font-family:Menlo,'DejaVu Sans Mono',consolas,'Courier New',monospace">{code}</pre>
    </code>
</body>
</html>
"""


def console_str(render_object: Any) -> "Text":
    console_str_callable = getattr(render_object, "__console_str__")
    if console_str_callable:
        return console_str_callable()
    else:
        return Text(str(render_object))


@dataclass
class ConsoleOptions:
    """Options for __console__ method."""

    min_width: int
    max_width: int
    is_terminal: bool
    encoding: str
    justify: Optional[JustifyValues] = None

    def copy(self) -> "ConsoleOptions":
        """Get a copy of this object.
        
        Returns:
            ConsoleOptions: New instance with same settings.
        """
        return replace(self)

    def update(
        self,
        width: int = None,
        min_width: int = None,
        max_width: int = None,
        justify: JustifyValues = None,
    ):
        """Update values, return a copy."""
        options = replace(self)
        if width is not None:
            options.min_width = options.max_width = width
        if min_width is not None:
            options.min_width = min_width
        if max_width is not None:
            options.max_width = max_width
        if justify is not None:
            options.justify = justify
        return options


@runtime_checkable
class ConsoleRenderable(Protocol):
    """An object that supports the console protocol."""

    def __console__(
        self, console: "Console", options: "ConsoleOptions"
    ) -> Iterable[Union["ConsoleRenderable", Segment]]:
        ...


RenderableType = Union[ConsoleRenderable, Segment, str]
RenderResult = Iterable[Union[ConsoleRenderable, Segment]]


class ConsoleDimensions(NamedTuple):
    """Size of the terminal."""

    width: int
    height: int


class StyleContext:
    """A context manager to manage a style."""

    def __init__(self, console: "Console", style: Optional[Style]):
        self.console = console
        self.style = style

    def __enter__(self) -> "Console":
        if self.style is not None:
            self.console.push_style(self.style)
        self.console._enter_buffer()
        return self.console

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.console._exit_buffer()
        if self.style is not None:
            self.console.pop_style()


COLOR_SYSTEMS = {
    "standard": ColorSystem.STANDARD,
    "256": ColorSystem.EIGHT_BIT,
    "truecolor": ColorSystem.TRUECOLOR,
}


class Console:
    """A high level console interface."""

    def __init__(
        self,
        color_system: Optional[
            Literal["auto", "standard", "256", "truecolor"]
        ] = "auto",
        styles: Dict[str, Style] = DEFAULT_STYLES,
        file: IO = None,
        width: int = None,
        height: int = None,
        record: bool = False,
        markup: bool = True,
        log_time: bool = True,
        log_path: bool = True,
        log_time_format: str = "[%X] ",
    ):

        self._styles = ChainMap(styles)
        self.file = file or sys.stdout
        self._width = width
        self._height = height
        self.record = record
        self._markup = markup

        if color_system is None:
            self._color_system = None
        elif color_system == "auto":
            self._color_system = self._detect_color_system()
        else:
            self._color_system = COLOR_SYSTEMS[color_system]

        self.buffer: List[Segment] = []
        self._buffer_index = 0
        self._record_buffer: List[Segment] = []

        default_style = Style()
        self.style_stack: List[Style] = [default_style]
        self.current_style = default_style

        self._log_render = LogRender(
            show_time=log_time, show_path=log_path, time_format=log_time_format
        )

    def __repr__(self) -> str:
        return f"<console width={self.width} {str(self._color_system)}>"

    def _detect_color_system(self,) -> Optional[ColorSystem]:
        """Detect color system from env vars."""
        if not self.is_terminal:
            return None
        if os.environ.get("COLORTERM", "").strip().lower() == "truecolor":
            return ColorSystem.TRUECOLOR
        # 256 can be considered standard nowadays
        return ColorSystem.EIGHT_BIT

    def _enter_buffer(self) -> None:
        """Enter in to a buffer context, and buffer all output."""
        self._buffer_index += 1

    def _exit_buffer(self) -> None:
        """Leave buffer context, and render content if required."""
        self._buffer_index -= 1
        self._check_buffer()

    def __enter__(self) -> "Console":
        """Own context manager to enter buffer context."""
        self._enter_buffer()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """Exit buffer context."""
        self._exit_buffer()

    def push_styles(self, styles: Dict[str, Style]) -> None:
        self._styles.maps.append(styles)

    @property
    def encoding(self) -> str:
        """Get the encoding of the console file.
        
        Returns:
            str: A standard encoding string.
        """
        return getattr(self.file, "encoding", "utf-8")

    @property
    def is_terminal(self) -> bool:
        """Check if the console is writing to a terminal.
        
        Returns:
            bool: True if the console writting to a device capable of
                understanding terminal codes, otherwise False.
        """
        isatty = getattr(self.file, "isatty", None)
        return False if isatty is None else isatty()

    @property
    def options(self) -> ConsoleOptions:
        """Get default console options."""
        return ConsoleOptions(
            min_width=0,
            max_width=self.width,
            encoding=self.encoding,
            is_terminal=self.is_terminal,
        )

    @property
    def size(self) -> ConsoleDimensions:
        """Get the size of the console.
        
        Returns:
            ConsoleDimensions: A named tuple containing the dimensions.
        """
        if self._width is not None and self._height is not None:
            return ConsoleDimensions(self._width, self._height)

        width, height = shutil.get_terminal_size()
        return ConsoleDimensions(
            width if self._width is None else self._width,
            height if self._height is None else self._height,
        )

    @property
    def width(self) -> int:
        """Get the width of the console.
        
        Returns:
            int: The width (in characters) of the console.
        """
        width, _ = self.size
        return width

    def line(self, count: int = 1) -> None:
        """Write new line(s).
        
        Args:
            count (int, optional): Number of new lines. Defaults to 1.
        """

        assert count >= 0, "count must be >= 0"
        if not count:
            return
        self.buffer.append(Segment("\n" * count))
        self._check_buffer()

    def _render(
        self, renderable: RenderableType, options: Optional[ConsoleOptions]
    ) -> Iterable[Segment]:
        """Render an object in to an iterable of `Segment` instances.

        This method contains the logic for rendering objects with the console protocol. 
        You are unlikely to need to use it directly, unless you are extending the library.

        
        Args:
            renderable (RenderableType): An object supporting the console protocol, or
                an object that may be converted to a string.
            options (ConsoleOptions, optional): An options objects. Defaults to None.
        
        Returns:
            Iterable[Segment]: An iterable of segments that may be rendered.
        """
        render_iterable: Iterable[RenderableType]
        render_options = options or self.options
        if isinstance(renderable, Segment):
            yield renderable
            return
        elif isinstance(renderable, ConsoleRenderable):
            render_iterable = renderable.__console__(self, render_options)
        elif isinstance(renderable, str):
            from .text import Text

            yield from self._render(Text(renderable), render_options)
            return
        else:
            raise errors.NotRenderableError(
                f"Unable to render {renderable!r}; "
                "A str, Segment or object with __console__ method is required"
            )

        for render_output in render_iterable:
            if isinstance(render_output, Segment):
                yield render_output
            else:
                yield from self.render(render_output, render_options)

    def render(
        self, renderable: RenderableType, options: Optional[ConsoleOptions]
    ) -> Iterable[Segment]:
        """Render an object in to an iterable of `Segment` instances.

        This method contains the logic for rendering objects with the console protocol. 
        You are unlikely to need to use it directly, unless you are extending the library.

        
        Args:
            renderable (RenderableType): An object supporting the console protocol, or
                an object that may be converted to a string.
            options (ConsoleOptions, optional): An options objects. Defaults to None.
        
        Returns:
            Iterable[Segment]: An iterable of segments that may be rendered.
        """
        yield from Segment.apply_style(
            self._render(renderable, options), self.current_style
        )

    def render_all(
        self, renderables: Iterable[RenderableType], options: Optional[ConsoleOptions]
    ) -> Iterable[Segment]:
        """Render a number of console objects.
        
        Args:
            renderables (Iterable[RenderableType]): Console objects.
            options (Optional[ConsoleOptions]): Options for render.
        
        Returns:
            Iterable[Segment]: Segments to be written to the console.
        
        """
        render_options = options or self.options
        for renderable in renderables:
            yield from self.render(renderable, render_options)

    def render_lines(
        self,
        renderable: RenderableType,
        options: Optional[ConsoleOptions],
        style: Optional[Style] = None,
    ) -> List[List[Segment]]:
        """Render objects in to a list of lines.

        The output of render_lines is useful when further formatting of rendered console text
        is required, such as the Panel class which draws a border around any renderable object.
        
        Args:
            renderables (Iterable[RenderableType]): Any object or objects renderable in the console.
            options (Optional[ConsoleOptions]): Console options used to render with.
        
        Returns:
            List[List[Segment]]: A list of lines, where a line is a list of Segment objects.
        """
        render_options = options or self.options

        with self.style(style or "none"):
            _rendered = self.render(renderable, render_options)
            lines = list(
                Segment.split_and_crop_lines(
                    _rendered, render_options.max_width, style=style
                )
            )
        return lines

    def render_str(self, text: str) -> "Text":
        """Convert a string to a Text instance.
        
        Args:
            text (str): Text to render.            
        
        Returns:
            ConsoleRenderable: Renderable object.
        
        """
        if self._markup:
            return markup.render(text)
        return Text(text)

    def _get_style(self, name: str) -> Optional[Style]:
        """Get a named style, or `None` if it doesn't exist.
        
        Args:
            name (str): The name of a style.
        
        Returns:
            Optional[Style]: A Style object for the given name, or `None`.
        """
        return self._styles.get(name, None)

    def get_style(
        self, name: Union[str, Style], *, default: Union[Style, str] = None
    ) -> Style:
        """Get a style merged with the current style.

        Args:
            name (str): The name of a style or a style definition.
        
        Returns:
            Style: A Style object.

        Raises:
            MissingStyle: If no style could be parsed from name.

        """
        if isinstance(name, Style):
            return name

        try:
            return self._styles.get(name) or Style.parse(name)
        except errors.StyleSyntaxError as error:
            if default is not None:
                return self.get_style(default)
            if " " in name:
                raise
            raise errors.MissingStyle(f"No style named {name!r}; {error}")

    def push_style(self, style: Union[str, Style]) -> None:
        """Push a style on to the stack.

        The new style will be applied to all `write` calls, until 
        `pop_style` is called.
        
        Args:
            style (Union[str, Style]): New style to merge with current style.
        
        Returns:
            None: [description]
        """
        if isinstance(style, str):
            style = self.get_style(style)
        self.current_style = self.current_style + style
        self.style_stack.append(self.current_style)

    def pop_style(self) -> Style:
        """Pop a style from the stack.

        This will revert to the style applied prior to the corresponding `push_style`.
        
        Returns:
            Style: The previously applied style.
        """
        if len(self.style_stack) == 1:
            raise errors.StyleStackError(
                "Can't pop the default style (check there is `push_style` for every `pop_style`)"
            )
        style = self.style_stack.pop()
        self.current_style = self.style_stack[-1]
        return style

    def style(self, style: Optional[Union[str, Style]]) -> StyleContext:
        """A context manager to apply a new style.

        Example:
            with context.style("bold red"):
                context.print("Danger Will Robinson!")
        
        Args:
            style (Union[str, Style]): New style to apply.
        
        Returns:
            StyleContext: A style context manager.
        """
        if style is None:
            return StyleContext(self, None)
        if isinstance(style, str):
            _style = self.get_style(style)
        else:
            if not isinstance(style, Style):
                raise TypeError(f"style must be a str or Style instance, not {style!r}")
            _style = style
        return StyleContext(self, _style)

    def _collect_renderables(
        self,
        objects: Iterable[Any],
        sep: str,
        end: str,
        emoji=True,
        highlight: "HighlighterType" = None,
    ) -> List[ConsoleRenderable]:
        """Combined a number of renderables and text in to one renderable.
        
        Args:
            renderables (Iterable[Union[str, ConsoleRenderable]]): [description]
            sep (str, optional): [description]. Defaults to " ".
        
        Returns:
            Renderables: [description]
        """
        from .text import Text

        sep_text = Text(sep)
        end_text = Text(end)
        renderables: List[ConsoleRenderable] = []
        append = renderables.append
        text: List[Text] = []
        append_text = text.append

        def check_text() -> None:
            if text:
                if end:
                    append_text(end_text)
                append(sep_text.join(text))
                del text[:]

        for renderable in objects:
            if isinstance(renderable, ConsoleRenderable):
                check_text()
                append(renderable)
                continue
            console_str_callable = getattr(renderable, "__console_str__", None)
            if console_str_callable is not None:
                append_text(console_str_callable())
                continue
            if isinstance(renderable, str):
                render_str = renderable
                if emoji:
                    render_str = _emoji_replace(render_str)
                render_text = self.render_str(render_str)
                append_text(highlight(render_text) if highlight else render_text)
            elif isinstance(renderable, Text):
                append_text(renderable)
            elif isinstance(renderable, (int, float, bool, bytes, type(None))):
                append_text(
                    highlight(repr(renderable)) if highlight else repr(renderable)
                )
            elif isinstance(renderable, (Mapping, Sequence)):
                check_text()
                append(Pretty(renderable))
            else:
                append_text(
                    highlight(repr(renderable)) if highlight else repr(renderable)
                )

        check_text()
        return renderables

    def print(
        self,
        *objects: Any,
        sep=" ",
        end="\n",
        style: Union[str, Style] = None,
        emoji=True,
        highlight: "HighlighterType" = None,
    ) -> None:
        """Print to the console.

        Args:
            *objects: Arbitrary objects to print to the console.
            sep (str, optional): Separator to print between objects. Defaults to " ".
            end (str, optional): Character to end print with. Defaults to "\n".
            style (Union[str, Style], optional):
            emoji (bool): If True, emoji codes will be replaced, otherwise emoji codes will be left in.
        """
        if not objects:
            self.line()
            return

        renderables = self._collect_renderables(
            objects, sep=sep, end=end, emoji=emoji, highlight=highlight
        )

        render_options = self.options
        extend = self.buffer.extend
        render = self.render
        with self.style(style):
            for renderable in renderables:
                extend(render(renderable, render_options))

    def log(
        self,
        *objects: Any,
        debug=Ellipsis,
        highlight: "HighlighterType" = None,
        log_locals: bool = False,
    ) -> None:
        if not objects and not debug:
            self.line()
            return
        highlighter = highlight or ReprHighlighter()
        renderables = self._collect_renderables(
            objects, sep=" ", end="\n", highlight=highlighter
        )

        if debug != Ellipsis:
            renderables.append(Pretty(debug))

        caller = inspect.stack()[1]
        path = caller.filename.rpartition(os.sep)[-1]
        line_no = caller.lineno
        if log_locals:
            locals_map = {
                key: value
                for key, value in caller.frame.f_locals.items()
                if not key.startswith("__")
            }
            renderables.append(tabulate_mapping(locals_map, title="Locals"))

        with self:
            self.buffer.extend(
                self.render(
                    self._log_render(self, renderables, path=path, line_no=line_no),
                    self.options,
                )
            )

    def _check_buffer(self) -> None:
        """Check if the buffer may be rendered."""
        if self._buffer_index == 0:
            text = self._render_buffer()
            self.file.write(text)

    def _render_buffer(self) -> str:
        """Render buffered output, and clear buffer."""
        output: List[str] = []
        append = output.append
        color_system = self._color_system
        buffer = self.buffer[:]
        if self.record:
            self._record_buffer.extend(buffer)
        del self.buffer[:]
        for line in Segment.split_and_crop_lines(buffer, self.width):
            for text, style in line:
                if style:
                    append(style.render(text, color_system=color_system, reset=True))
                else:
                    append(text)
            append("\n")
        rendered = "".join(output)
        return rendered

    def export_text(self, clear: bool = True, styles: bool = False) -> str:
        """Generate text from console contents (requires record=True argument in constructor).
        
        Args:         
            clear (bool, optional): Set to True to clear the record buffer after exporting.
            styles (bool, optional): If True, ansi style codes will be included. False for plain text.
                Defaults to False.
                   
        Returns:
            str: String containing console contents.

        """
        assert (
            self.record
        ), "To export console contents set record=True in the constructor or instance"
        if styles:
            text = "".join(
                (style.render(text, reset=True) if style else text)
                for text, style in self._record_buffer
            )
        else:
            text = "".join(text for text, _ in self._record_buffer)
        if clear:
            del self._record_buffer[:]
        return text

    def save_text(self, path: str, clear: bool = True, styles: bool = False) -> None:
        """Generate text from console and save to a given location (requires record=True argument in constructor).
        
        Args:   
            path (str): Path to write text files.                    
            clear (bool, optional): Set to True to clear the record buffer after exporting.
            styles (bool, optional): If True, ansi style codes will be included. False for plain text.
                Defaults to False.     

        """
        text = self.export_text(clear=clear, styles=styles)
        with open(path, "wt") as write_file:
            write_file.write(text)

    def export_html(
        self,
        theme: Theme = None,
        clear: bool = True,
        code_format=CONSOLE_HTML_FORMAT,
        inline_styles: bool = False,
    ) -> str:
        """Generate HTML from console contents (requires record=True argument in constructor).
        
        Args:            
            theme (Theme, optional): Theme object containing console colors.
            clear (bool, optional): Set to True to clear the record buffer after generating the HTML.
            code_format (str, optional): Format string to render HTML, should contain {foreground}
                {background} and {code}.
            inline_styes (bool, optional): If True styles will be inlined in to spans, which makes files
                larger but easier to cut and paste markup. If False, styles will be embedded in a style tag.
                Defaults to False.
        
        Returns:
            str: String containing console contents.
        """
        assert (
            self.record
        ), "To export console contents set record=True in the constructor or instance"
        fragments: List[str] = []
        append = fragments.append
        _theme = theme or themes.DEFAULT
        stylesheet = ""

        if inline_styles:
            for text, style in Segment.simplify(self._record_buffer):
                if style:
                    rule = style.get_html_style(_theme)
                    append(f'<span style="{rule}">{text}</span>' if rule else text)
                else:
                    append(text)
        else:
            styles: Dict[str, int] = {}
            for text, style in Segment.simplify(self._record_buffer):
                if style:
                    rule = style.get_html_style(_theme)
                    if rule:
                        style_number = styles.setdefault(rule, len(styles) + 1)
                        append(f'<span class="r{style_number}">{text}</span>')
                    else:
                        append(text)
                else:
                    append(text)
            stylesheet_rules: List[str] = []
            stylesheet_append = stylesheet_rules.append
            for style_rule, style_number in styles.items():
                if style_rule:
                    stylesheet_append(f".r{style_number} {{{style_rule}}}")
            stylesheet = "\n".join(stylesheet_rules)

        rendered_code = code_format.format(
            code="".join(fragments),
            stylesheet=stylesheet,
            foreground=_theme.foreground_color.hex,
            background=_theme.background_color.hex,
        )
        if clear:
            del self._record_buffer[:]
        return rendered_code

    def save_html(
        self,
        path: str,
        theme: Theme = None,
        clear: bool = True,
        code_format=CONSOLE_HTML_FORMAT,
        inline_styles: bool = False,
    ) -> None:
        """Generate HTML from console contents and write to a file (requires record=True argument in constructor).
        
        Args:  
            path (str): Path to write html file.        
            theme (Theme, optional): Theme object containing console colors.
            clear (bool, optional): Set to True to clear the record buffer after generating the HTML.
            code_format (str, optional): Format string to render HTML, should contain {foreground}
                {background} and {code}.
            inline_styes (bool, optional): If True styles will be inlined in to spans, which makes files
                larger but easier to cut and paste markup. If False, styles will be embedded in a style tag.
                Defaults to False.
        
        """
        html = self.export_html(
            theme=theme,
            clear=clear,
            code_format=code_format,
            inline_styles=inline_styles,
        )
        with open(path, "wt") as write_file:
            write_file.write(html)


if __name__ == "__main__":  # pragma: no cover
    console = Console()

    with console.style("dim on black"):
        console.print("[b]Hello[/b], [i]World[/i]!")
        console.print("Hello, *World*!")

    console.log(
        "JSONRPC *request*",
        5,
        1.3,
        True,
        False,
        None,
        {
            "jsonrpc": "2.0",
            "method": "subtract",
            "params": {"minuend": 42, "subtrahend": 23},
            "id": 3,
        },
    )

    console.log("# Hello, **World**!")
    console.log("Hello, World!", "{'a': 1}", repr(console))

    console.log(
        {
            "name": None,
            "empty": [],
            "quiz": {
                "sport": {
                    "answered": True,
                    "q1": {
                        "question": "Which one is correct team name in NBA?",
                        "options": [
                            "New York Bulls",
                            "Los Angeles Kings",
                            "Golden State Warriros",
                            "Huston Rocket",
                        ],
                        "answer": "Huston Rocket",
                    },
                },
                "maths": {
                    "answered": False,
                    "q1": {
                        "question": "5 + 7 = ?",
                        "options": [10, 11, 12, 13],
                        "answer": 12,
                    },
                    "q2": {
                        "question": "12 - 8 = ?",
                        "options": [1, 2, 3, 4],
                        "answer": 4,
                    },
                },
            },
        }
    )
# console.print("foo")




from typing import Iterator, Iterable, List, TypeVar, TYPE_CHECKING, Union
from typing_extensions import Literal

from .segment import Segment


if TYPE_CHECKING:
    from .console import (
        Console,
        ConsoleOptions,
        ConsoleRenderable,
        RenderResult,
        RenderableType,
    )
    from .text import Text

from ._render_width import RenderWidth

T = TypeVar("T")


class Renderables:
    """A list subclass which renders its contents to the console."""

    def __init__(self, renderables: Iterable["ConsoleRenderable"] = None) -> None:
        self._renderables: List["ConsoleRenderable"] = (
            list(renderables) if renderables is not None else []
        )

    def __console__(
        self, console: "Console", options: "ConsoleOptions"
    ) -> "RenderResult":
        """Console render method to insert line-breaks."""
        yield from self._renderables

    def __console_width__(self, max_width: int) -> "RenderWidth":
        dimensions = [
            RenderWidth.get(renderable, max_width) for renderable in self._renderables
        ]
        _min = max(dimension.minimum for dimension in dimensions)
        _max = max(dimension.maximum for dimension in dimensions)
        return RenderWidth(_min, _max)

    def append(self, renderable: "ConsoleRenderable") -> None:
        self._renderables.append(renderable)

    def __iter__(self) -> Iterable["ConsoleRenderable"]:
        return iter(self._renderables)


class Lines:
    """A list subclass which can render to the console."""

    def __init__(self, lines: Iterable["Text"] = ()) -> None:
        self._lines: List["Text"] = list(lines)

    def __iter__(self) -> Iterator["Text"]:
        return iter(self._lines)

    def __getitem__(self, index: int) -> "Text":
        return self._lines[index]

    def __setitem__(self, index: int, value: "Text") -> "Lines":
        self._lines[index] = value
        return self

    def __len__(self) -> int:
        return self._lines.__len__()

    def __console__(
        self, console: "Console", options: "ConsoleOptions"
    ) -> "RenderResult":
        """Console render method to insert line-breaks."""
        for line in self._lines:
            yield line

    def append(self, line: "Text") -> None:
        self._lines.append(line)

    def extend(self, lines: Iterable["Text"]) -> None:
        self._lines.extend(lines)

    def justify(
        self,
        width: int,
        align: Literal["none", "left", "center", "right", "full"] = "left",
    ) -> None:
        """Pad each line with spaces to a given width.
        
        Args:
            width (int): Number of characters per line.
            
        """
        from .text import Text

        if align == "left":
            for line in self._lines:
                line.pad_right(width - len(line.text))
        elif align == "center":
            for line in self._lines:
                line.pad_left((width - len(line.text)) // 2)
                line.pad_right(width - len(line.text))
        elif align == "right":
            for line in self._lines:
                line.pad_left(width - len(line.text))
        elif align == "full":
            for line_index, line in enumerate(self._lines):
                if line_index == len(self._lines) - 1:
                    break
                words = line.split(" ")
                words_size = sum(len(word) for word in words)
                num_spaces = len(words) - 1
                spaces = [1 for _ in range(num_spaces)]
                index = 0
                while words_size + num_spaces < width:
                    spaces[len(spaces) - index - 1] += 1
                    num_spaces += 1
                    index = (index + 1) % len(spaces)
                tokens: List[Text] = []
                index = 0
                for index, word in enumerate(words):
                    tokens.append(word)
                    if index < len(spaces):
                        tokens.append(Text(" " * spaces[index]))
                    index += 1
                self[line_index] = Text("").join(tokens)
